% filter design example using filter design toolbox of MATLAB
% (as an example of an object-oriented programming application)
%
% G.Doblinger, TU-Wien, 11-2009

% bandpass filter specifications

%                ________  
% Ap             ________
%               |        |
%
% As1_______|                |_______As2
%   0       fs1 fp1      fp2 fs2     Fs/2

clear all
clc

Fs = 16000;   % sampling frequency in Hz
fs1 = 1800;   % lower stopband frequency in Hz
fp1 = 2000;   % lower passband frequency in Hz
fp2 = 3000;   % upper passband frequency in Hz
fs2 = 3200;   % upper stopband frequency in Hz
As1 = 80;     % lower stopband attenuation in dB
Ap = 1.0;     % passband ripple in dB
As2 = 80;     % upper stopband attenuation in dB

% call filter design class constructor of bandpass filters

D = fdesign.bandpass('Fst1,Fp1,Fp2,Fst2,Ast1,Ap,Ast2',fs1,fp1,fp2,fs2,...
      As1,Ap,As1,Fs);

disp(D);            % show bandpass class properties

designmethods(D)    % show possible desing methods

%%

% first design an equiripple FIR filter
% Note: plot of impulse response, phase response etc. can be selected
% in plot window of fvtool

Hfir = design(D,'equiripple'); 

disp(Hfir);

close all;
fvtool(Hfir);       % filter visualization tool (plot magnitude response)
impz(Hfir);         % plot impulse response  
grpdelay(Hfir);     % plot group delay (phase differentiated by frequency)

fprintf(1,'FIR filter length = %d\n\n', length(Hfir.Numerator));

%%

% repeat with elliptic IIR filter
% structure: cascade of 2nd order sections (direct form I)

Hiir1 = design(D,'ellip','FilterStructure','df1sos'); 

disp(Hiir1);

close all;
fvtool(Hiir1);

Nimp = 400;         % number of impulse response samples
impz(Hiir1,Nimp);
grpdelay(Hiir1);

fprintf(1,'number of 2nd order sections = %d\n\n', size(Hiir1.sosMatrix,1));

%%

% quantize FIR filter coefficients

Nw = 12;             % word length in bits
Hfirq = Hfir;
Hfirq.Arithmetic = 'fixed';
set(Hfirq, 'CoeffWordLength',Nw);

close all;
fvtool(Hfirq);

%%

% repeat with IIR filter

Nw = 12;              % word length in bits
Hiirq = Hiir1;
Hiirq.Arithmetic = 'fixed';
set(Hiirq, 'CoeffWordLength',Nw);
fvtool(Hiirq);

%%

% use direct form II 2nd order sections

Hiir2 = design(D,'ellip','FilterStructure','df2sos'); 
Nw = 12;              % word length in bits
Hiirq = Hiir2;
Hiirq.Arithmetic = 'fixed';
set(Hiirq, 'CoeffWordLength',Nw);
fvtool(Hiirq);


