function D = CompleteTree(x,LoF_D,HiF_D,N)
%function D = CompleteTree(x,LoF_D,HiF_D,N)
%
% compute full wavelet packet binary tree and shannon entropy
%
% x            input signal vector
% LoF_D,HiF_D  decomposition filter coefficient vectors
% N            number of tree levels for signal decomposition (N >= 1)
% D            structure of node data
%              D.coef     node coefficients concatenated to row vector
%              D.start    start index of coeff. for each node
%              D.stop     stop index of coeff. for each node
%              D.root     index of root node for each node (top node = index 0)
%              D.entr     entropy for each node
%              D.engy     energy for each node

% G.Doblinger, TU-Wien, 7.12.97

Nnodes = 2^(N+1)-2;     % number of nodes in the full binary tree (without first node)

% initialize node data structure

D = struct('coef',[],'start',zeros(1,Nnodes),'stop',zeros(1,Nnodes),'root',...
   zeros(1,Nnodes),'entr',zeros(1,Nnodes),'engy',zeros(1,Nnodes));
L = length(x);
Lh = length(LoF_D);
nstart = 1;
n = 1;
for k = 1:N
   L = floor((L+Lh-1)/2);      % length of coefficients
   for l = 1:2^k
      D.start(n) = nstart;
      nstop = nstart + L - 1;
      D.stop(n) = nstop;
      D.root(n) = floor((n-1)/2);
      nstart = nstop + 1;
      n = n + 1;
   end
end
D.coef = zeros(1,D.stop(Nnodes));

% compute first 2 node coefficients

y = ConvDeci(x,LoF_D);
D.coef(D.start(1):D.stop(1)) = y;
D.entr(1) = ShanEntr(y);
D.engy(1) = sum(y.^2);
y = ConvDeci(x,HiF_D);
D.coef(D.start(2):D.stop(2)) = y;
D.entr(2) = ShanEntr(y);
D.engy(2) = sum(y.^2);

% compute remaining nodes

for n = 3:2:Nnodes-1
   x_node = D.coef(D.start(D.root(n)):D.stop(D.root(n)));   % get root coeff.
   y = ConvDeci(x_node,LoF_D);
   D.coef(D.start(n):D.stop(n)) = y;
   D.entr(n) = ShanEntr(y);
   D.engy(n) = sum(y.^2);
   y = ConvDeci(x_node,HiF_D);
   D.coef(D.start(n+1):D.stop(n+1)) = y;
   D.entr(n+1) = ShanEntr(y);
   D.engy(n+1) = sum(y.^2);
end
