function D = CompleteTree1(x,LoF_D,HiF_D,N)
%function D = CompleteTree1(x,LoF_D,HiF_D,N)
%
% compute full wavelet packet binary tree and shannon entropy
% (version with spectrum inversion of high pass section to correct inversion due to
% downsapmling by 2, and coefficient delay for each node)
%
% x            input signal vector
% LoF_D,HiF_D  decomposition filter coefficient vectors (same filter length required)
% N            number of tree levels for signal decomposition (N >= 1)
% D            structure of node data
%              D.coef     node coefficients concatenated to row vector
%              D.start    start index of coeff. for each node
%              D.stop     stop index of coeff. for each node
%              D.root     index of root node for each node (top node = index 0)
%              D.entr     entropy for each node
%              D.engy     energy for each node
%              D.delay    coefficient delay for each node

% G. Doblinger, TU-Wien, 7.12.97

L = length(x);
if N == 0
   D = struct('coef',x,'start',1,'stop',L,'root',1,'entr',ShanEntr(x),'engy',sum(x.^2),...
      'delay',0);
   return
end

Nnodes = 2^(N+1)-2;     % number of nodes in the full binary tree (without first node)

% compute energy delays of filters (used for delay compensation)
% Note: delay of impulse response maximum is computed for each node
% using the concolution/decimation formula, the delay of the maximun can be
% easily computed by delay(n) = (delay(root node) + Nd)/2 where Nd is the respective
% filter delay.
% Attention: this method works only for non-minimum Phase filters, where the maximum
% is shifted towards the origin by decimation.

Lh = length(LoF_D);
Nd_lp = sum((0:Lh-1)'.*(LoF_D(:).^2))/sum(LoF_D.^2);
Nd_hp = sum((0:Lh-1)'.*(HiF_D(:).^2))/sum(HiF_D.^2);
if Nd_lp < 5 || Nd_hp < 5
   disp('INFO: low-delay analysis filters detected, delay compensation may fail!');
end

% initialize node data structure

D = struct('coef',[],'start',zeros(1,Nnodes),'stop',zeros(1,Nnodes),'root',...
   zeros(1,Nnodes),'entr',zeros(1,Nnodes),'engy',zeros(1,Nnodes),'delay',zeros(1,Nnodes));
nstart = 1;
n = 1;
for k = 1:N
   L = floor((L+Lh-1)/2);      % length of coefficients
   for l = 1:2^k
      D.start(n) = nstart;
      nstop = nstart + L - 1;
      D.stop(n) = nstop;
      D.root(n) = floor((n-1)/2);
      nstart = nstop + 1;
      n = n + 1;
   end
end
D.coef = zeros(1,D.stop(Nnodes));

% compute first 2 node coefficients

y = ConvDeci(x,LoF_D);
D.coef(D.start(1):D.stop(1)) = y;
D.entr(1) = ShanEntr(y);
D.engy(1) = sum(y.^2);
D.delay(1) = Nd_lp/2;
y = ConvDeci(x,HiF_D);
nodd = 2:2:length(y);
y(nodd) = -y(nodd);            % invert spectrum of high pass output
D.coef(D.start(2):D.stop(2)) = y;
D.entr(2) = ShanEntr(y);
D.engy(2) = sum(y.^2);
D.delay(2) = Nd_hp/2;

% compute remaining nodes

for n = 3:2:Nnodes-1
   x_node = D.coef(D.start(D.root(n)):D.stop(D.root(n)));   % get root coeff.
   y = ConvDeci(x_node,LoF_D);
   D.coef(D.start(n):D.stop(n)) = y;
   D.entr(n) = ShanEntr(y);
   D.engy(n) = sum(y.^2);
   y = ConvDeci(x_node,HiF_D);
   nodd = 2:2:length(y);
   y(nodd) = -y(nodd);                                      % invert high pass spectrum
   D.coef(D.start(n+1):D.stop(n+1)) = y;
   D.entr(n+1) = ShanEntr(y);
   D.engy(n+1) = sum(y.^2);
   Del_root = D.delay(D.root(n));                           % delay of root node
   D.delay(n) = (Del_root + Nd_lp)/2;                       % low pass path delay
   D.delay(n+1) = (Del_root + Nd_hp)/2;                     % high pass path delay
end
D.delay = round(D.delay);
