function [xe,xn,D] = wpdn_fir(x,Nblk,Nov,Novs,SNR,N,wname,treetype,SorH)
% function [xe,xn,D] = wpdn_fir(x,Nblk,Nov,Novs,SNR,N,wname,treetype,SorH)
%
% signal de-noising using wavelet packets (block processing, automatic threshold)
% (prototype for DSP implementation)
% (energy of node signals is used to find approximation nodes)
%
% x         clean signal vector
% Nblk      block size Nblk <= length(x)
% Nov       block overlapping in samples (for analysis)
% Novs      block overlapping for reconstruction (Novs <= Nov)
% SNR       SNR of noisy signal
% N         number of analysis levels
% wname     wavelet type in quotes ('coif4' typ.)
%           or filter vector h designed with qmf_2.m or qmf_2hb.m
% treetype  'b' for best basis tree, 'f' for full binary tree, 'w' for wavelet tree
% SorH      's' for soft, 'h' for hard thresholding (default 's')
% xe        enhanced signal vector
% xn        noisy signal vector
% D         structure of node data (for last signal block)
%           D.coef            node coefficients concatenated to row vector
%           D.start, D.stop   start, stop index of coeff. of each node
%           D.root            index of root node of each node (top node = index 0)
%           D.entr            entropy of each node (NaN, if node is omitted)
%           D.engy            energy of each node
%
% example: [xar,a] = armod(1000,[0.98 0.97],[0.04 0.07],1);
%          [xare,xn,D] = wpdn_fir(xar,200,20,4,10,3,'coif4','b',123);

% G. Doblinger TU-Wien, 11-97, 11-2009

if nargin == 0
   help wpdn_fir
   return
elseif nargin == 8
   SorH = 's';
end
if N <= 0
   error('number of tree levels must be greater than zero');
elseif Nov >= Nblk
   error('overlapping Nov must be less than block size Nblk');
elseif Novs > Nov
   error('overlapping Novs must be less than or equal to Nov');
end

% make lengths even

Ndata = length(x);
if mod(Ndata,2)
   Ndata = Ndata-1;
   disp('signal length set to even number');
end
if mod(Nblk,2)
   Nblk = Nblk-1;
   disp('Nblk set to even number');
end
if mod(Nov,2)
   Nov = Nov-1;
   disp('Nov set to even number');
end
if mod(Novs,2)
   Novs = Novs-1;
   disp('Novs set to even number');
end
if Nblk > Ndata
   Nblk = Ndata;
end

x = x(:);
x = x - mean(x);

% add white noise

if SNR ~= inf
   v = randn(size(x));
   v = v - mean(v);
   sigma_x = std(x);
   if sigma_x <= eps                     % DC signal
      sigma_x = mean(x) + eps;
   end
   v = (sigma_x/std(v))*10^(-SNR/20)*v;
   xn = x + v;                           % noisy signal
else
   xn = x;
   v = zeros(size(x));
end

keepapp = 1;   % do not change approximation part (low pass filter) coefficients

% compute global threshold estimate from noisy signal

sigma_v = std(v);
threshold = sigma_v*sqrt(2*log(Ndata));

% compute wavelet decomposition and reconstruction filters

if ischar(wname)
   [LoF_D,HiF_D,LoF_R,HiF_R] = wfilters(wname);
else
   Lh = round(length(wname)/4);
   LoF_D = wname(1:Lh);
   if abs(sum(LoF_D)-sqrt(2)) > 1e-3
      error('no valied filter vector');
   end
   HiF_D = wname(Lh+1:2*Lh);
   LoF_R = wname(2*Lh+1:3*Lh);
   HiF_R = wname(3*Lh+1:4*Lh);
end

% initalize block processing

if Nblk == Ndata                         % only a single block
   Nblk1 = Nblk;
   Nblk2 = Nblk;
   dNov = 0;
   M = 0;
   Ndat = Ndata;
else
   Nblk1 = Nblk-Nov;
   Nblk2 = Nblk-Novs;
   dNov = Nov - Novs;
   M = floor((Ndata-Nblk)/Nblk1);
   Ndat = Nblk+M*Nblk1;
end
xe = zeros(Ndat,1);
xn = xn(1:Ndat);
Ntn = zeros(1,M+1);
treetype = lower(treetype);

for m = 1:M+1

   nstart_xn = (m-1)*Nblk1+1;            % advance blocks by Nblk1 samples
   xt = xn(nstart_xn:nstart_xn+Nblk-1);  % extract signal block of length Nblk
                                         % overlapping of analysis blocks = Nov

   % compute wavelet packet decomposition, find and best tree

   D = CompleteTree(xt,LoF_D,HiF_D,N);   % compute node coefficients of full binary
                                         % tree
   if treetype == 'b'
      D = BestBasisTree(D);              % find best basis tree structure
   elseif treetype == 'w'
      D = WaveletTree(D);
   end
   Ntn(m) = length(TermNodes(D));        % get number of terminal nodes

   % wavelet packet coefficients thresholding

   D = ThresCoef(D,keepapp,SorH,threshold);

   % wavelet packet reconstruction to form de-noised signal
   % (overlap-save method, overlapping of reconstruction blocks = Novs < Nov
   % to avoid boundary effects)

   xd = WavePackRec(D,LoF_R,HiF_R);
   if m == 1
      xe(1:Nblk2) = xd(1:Nblk2);
      nstart_xe = Nblk2+1;
   else
      nstop_xe = nstart_xe+Nblk1-1;
      xe(nstart_xe:nstop_xe) = xd(dNov+1:Nblk2);
      nstart_xe = nstop_xe+1;
   end
end

Ndat = Ndat - Novs;
xe = xe(1:Ndat);
x = x(1:Ndat);
xn = xn(1:Ndat);
[nr,nc] = size(x);
if nr > nc
   xe = xe(:);
else
   xe = xe(:).';
end

% plot results

close all
upperright = [0.5,0.5,0.49,0.45];
figure('name','Wavelet packet de-noising','Units','normal',...
   'Position',upperright);

n = (0:Ndat-1)';
ystr1 = 'Amplitude';
tstr1 = ['Noisy signal (SNR = ', num2str(SNR), ' dB, No. of levels = ', int2str(N),...
        ', threshold = ', num2str(threshold), ')'];
xstr2 = 'Time index n';
tstr2 =  'Clean signal and its estimate';
ymax = max([abs(x) ; abs(xn)]);
subplot(2,1,1), plot(n,xn,'b'), ylabel(ystr1), title(tstr1), grid on
axis([0 Ndat-1 -ymax ymax]);
subplot(2,1,2), plot(n,x,'b',n,xe,'r'), xlabel(xstr2), ylabel(ystr1), title(tstr2), grid on
axis([0 Ndat-1 -ymax ymax]);
legend('clean','denoised');

e_est = x-xe;
R_out = norm(xe)/norm(e_est);
if SNR ~= inf
   SNR_e = 20*log10(R_out)-SNR;
   disp(['Estimated SNR improvement in dB: ', num2str(SNR_e)]);
   SNR_d = -20*log10(norm(e_est)/norm(x));
   disp(['Estimated signal distortion in dB: ', num2str(SNR_d)]);
end

% plot wavelet tree

if M == 0   % only one block
   disp(['Number of terminal nodes: ',int2str(Ntn)]);
   pos = [0.005,0.5,0.49,0.45];
   term_nodes = TermNodes(D);                % get terminal nodes
   Ntn = length(term_nodes);
   if Ntn > 1
      PlotWavTree(D,pos,'Signal decomposition tree');
   end
else
   if treetype == 'b'
      lowerright = [0.5,0.04,0.49,0.4];
      figure('name','Tree complexity','Units','normal','Position',lowerright);
      m = 0:Nblk1:M*Nblk1;
      bar(m,Ntn,0.7); xlabel(xstr2), title('Number of terminal nodes per block')
      set(gca,'XLim',[0,inf],'YTick',(0:1:max(Ntn)));
   else
      disp(['Number of terminal nodes: ',int2str(Ntn(1))]);
   end
end

